// SPDX-License-Identifier: CC0-1.0

// Copyright (C) 2021 Wojtek Kosior <koszko@koszko.org>
//
// Available under the terms of Creative Commons Zero v1.0 Universal.

var hello_message = "hello, "
var bye_message = "bye, "
